﻿using System;
using System.Collections.Generic;
using System.Text;

namespace Dámy
{
    public class Řešitel
    {
        private readonly int početDam;
        private readonly int[] řešení;
        public int PočetŘešení { get; private set; } = 0;
        public IEnumerable<int[]> Řešení
        {
            get
            {
                while (NajdiDalšíŘešení(out int[] p))
                {
                    yield return p;
                }
            }
        }

        public bool NajdiDalšíŘešení(out int[] výsledek)
        {
            výsledek = PočetŘešení == 0 ? ŘešOdSloupce(0) : ŘešOdSloupce(početDam - 1);
            if (výsledek != null)
            {
                PočetŘešení++;
            }

            return výsledek != null;
        }

        public void NajdiVšechnaŘešení()
        {
            //NajdiŘešení(0);
            //int[] p = null;
            while (NajdiDalšíŘešení(out int[] p))
                ;
        }



        private bool Bezpečné(int k, int m)// k-tá řádka, m-tý sloupec
        {
            if (m == 0) return true;    // V prvním sloupci lze umístit kamkoli
            var (výsledek, j) = (true, 0);
            while (výsledek && (j < m))         // Projdi všechny předchozí sloupce
            {
                výsledek = (řešení[j] != k) &&      // stejná řádka?
                    (řešení[m - j - 1] != k - j - 1) && // stejná diagonála?
                    (řešení[m - j - 1] != k + j + 1);
                if (!výsledek) return false;
                j++;
            }
            return true;
        }

        public void Výpis()
        {
            Console.Write("(");
            for (int i = 0; i < početDam; i++)
            {
                Console.Write(řešení[i] + (i < početDam - 1 ? ", " : ""));
            }
            Console.WriteLine(")");
        }

        private int NajdiDalšíBezpečnouŘádku(int k, int m)   // Najde další možné umístění v m-tém sloupci, začne za k-tou řádkou
        {
            while (++k < početDam)
            {
                if (Bezpečné(k, m))
                {
                    break;
                }
            }
            return k;                           // Neuspěje-li, vrátí početDam
        }

        private int[] ŘešOdSloupce(int m)            // Začne hledat další řešení, počínaje m-tým sloupcem
        {
            int k;
            while (m > -1)
            {
                k = NajdiDalšíBezpečnouŘádku(řešení[m], m);
                if (k < početDam)
                {
                    řešení[m++] = k;
                    if (m == početDam) return řešení;
                }
                else
                {
                    řešení[m--] = -1;
                }
            }
            return null;
        }

        public void NajdiŘešení(int m)
        {
            if (m == početDam)
            {
                //Výpis();
                PočetŘešení++;
            }
            else
            {
                for (int k = 0; k < početDam; k++)
                {
                    if (Bezpečné(k, m))
                    {
                        řešení[m] = k;
                        NajdiŘešení(m + 1);
                    }
                }
            }
            if (m < početDam)
            {
                řešení[m] = -1;
            }
        }

        public bool NajdiDalšíŘešeníOdŘádky(int řádka)
        {
            if (řešení[0] == -1) řešení[0] = řádka;
            int[] výsledek = PočetŘešení == 0 ? ŘešOdSloupce(1) : ŘešOdSloupce(početDam - 1);
            if (výsledek != null && řešení[0] == řádka)
            {
                PočetŘešení++;
            }
            return výsledek != null && řešení[0] == řádka;
        }


        public Řešitel(int n)           // Konstruktor řešitele pro úlohu n dam
        {
            if(n < 4)
            {
                throw new ArgumentOutOfRangeException("Počet dam musí být alespoň 4");
            }
            this.početDam = n;
            řešení = new int[n];
            for (int i = 0; i < n; i++)
            {
                řešení[i] = -1;
            }
        }
    }

}
